@@ -17,17 +17,44 @@ else:
     unicode_type = unicode
 unicode_dtype = np.dtype(unicode_type)
 
+def check_dataframe_deps():
+    pd_error = """Pandas version >= 1.0 required for dataframe functionality.
+                  Please `pip install pandas>=1.0` to proceed."""
+    pa_error = """PyArrow version >= 1.0 is suggested for dataframe functionality.
+                  Please `pip install pyarrow>=1.0`."""
+
+    from distutils.version import LooseVersion
+
+    try:
+        import pandas as pd
+    except ImportError:
+        raise Exception(pd_error)
+
+    if LooseVersion(pd.__version__) < LooseVersion("1.0"):
+        raise Exception(pd_error)
+
+    try:
+        import pyarrow as pa
+        if LooseVersion(pa.__version__) < LooseVersion("1.0"):
+            warnings.warn(pa_error)
+    except ImportError:
+        warnings.warn(pa_error)
+
 # TODO
 # - handle missing values
 # - handle extended datatypes
 
+# Note: 'None' is used to indicate optionality for many of these options
+#       For example, if the `sparse` argument is unspecified we will default
+#       to False (dense) unless the input has string or heterogenous indexes.
 TILEDB_KWARG_DEFAULTS = {
     'ctx': None,
-    'sparse': True,
+    'sparse': None,
     'index_dims': None,
     'allows_duplicates': True,
     'mode': 'ingest',
-    'attrs_filters': None,
+    'attr_filters': None,
+    'dim_filters': None,
     'coords_filters': None,
     'full_domain': False,
     'tile': None,
@@ -117,6 +144,17 @@ def attrs_from_df(df,
 
     attrs = list()
     for name, col in df.items():
+        if isinstance(filters, dict):
+            if name in filters:
+                attr_filters = filters[name]
+            else:
+                attr_filters = None
+        elif filters is not None:
+            attr_filters = filters
+        else:
+            attr_filters = tiledb.FilterList(
+                [tiledb.ZstdFilter(1, ctx=ctx)])
+
         # ignore any column used as a dim/index
         if index_dims and name in index_dims:
             continue
@@ -129,7 +167,9 @@ def attrs_from_df(df,
             attr_info = ColumnInfo(spec_type)
         else:
             attr_info = dtype_from_column(col)
-        attrs.append(tiledb.Attr(name=name, dtype=attr_info.dtype, filters=filters))
+
+        attrs.append(tiledb.Attr(name=name, dtype=attr_info.dtype,
+                                 filters=attr_filters))
 
         if attr_info.repr is not None:
             attr_reprs[name] = attr_info.repr
@@ -160,7 +200,8 @@ def dim_info_for_column(ctx, df, col, tile=None, full_domain=False, index_dtype=
 
     return dim_info
 
-def dim_for_column(ctx, name, dim_info, col, tile=None, full_domain=False, ndim=None):
+def dim_for_column(ctx, name, dim_info, col, tile=None, full_domain=False,
+                   ndim=None, dim_filters=None):
     if isinstance(col, np.ndarray):
         col_values = col
     else:
@@ -212,8 +253,11 @@ def dim_for_column(ctx, name, dim_info, col, tile=None, full_domain=False, ndim=
     if not dim_info.dtype in (np.bytes_, np.unicode):
         if np.issubdtype(dtype, np.integer):
             dim_range = np.uint64(np.abs(np.uint64(dim_max) - np.uint64(dim_min)))
+            # we can't make a tile larger than the dimension range
             if dim_range < tile:
                 tile = dim_range
+            if tile < 1:
+                tile = 1
         elif np.issubdtype(dtype, np.float64):
             dim_range = dim_max - dim_min
             if dim_range < tile:
@@ -223,7 +267,8 @@ def dim_for_column(ctx, name, dim_info, col, tile=None, full_domain=False, ndim=
         name = name,
         domain = (dim_min, dim_max),
         dtype = dim_info.dtype,
-        tile = tile
+        tile = tile,
+        filters=dim_filters
     )
 
     return dim
@@ -240,7 +285,7 @@ def get_index_metadata(dataframe):
     return md
 
 def create_dims(ctx, dataframe, index_dims,
-                tile=None, full_domain=False, sparse=None):
+                tile=None, full_domain=False, sparse=None, filters=None):
     import pandas as pd
     index = dataframe.index
     index_dict = OrderedDict()
@@ -290,6 +335,7 @@ def create_dims(ctx, dataframe, index_dims,
                          tile=dim_tile, full_domain=full_domain,
                          index_dtype=index_dtype))
 
+
     if any([d.dtype in (np.bytes_, np.unicode_) for d in dim_types]):
         if sparse is False:
             raise TileDBError("Cannot create dense array with string-typed dimensions")
@@ -303,10 +349,25 @@ def create_dims(ctx, dataframe, index_dims,
         elif sparse is None:
             sparse = True
 
+    # Fall back to default dense type if unspecified and not inferred from dimension types
+    if sparse is None:
+        sparse = False
+
     ndim = len(dim_types)
 
     dims = list()
     for idx, (name, values) in enumerate(index_dict.items()):
+        # get the FilterList, if any
+        if isinstance(filters, dict):
+            if name in filters:
+                dim_filters = filters[name]
+            else:
+                dim_filters = None
+        elif filters is not None:
+            dim_filters = filters
+        else:
+            dim_filters = None
+
         if per_dim_tile and name in tile:
             dim_tile = tile[name]
         elif per_dim_tile:
@@ -317,7 +378,8 @@ def create_dims(ctx, dataframe, index_dims,
             dim_tile = tile
 
         dims.append(dim_for_column(ctx, name, dim_types[idx], values,
-                    tile=dim_tile, full_domain=full_domain, ndim=ndim))
+                    tile=dim_tile, full_domain=full_domain, ndim=ndim,
+                    dim_filters=dim_filters))
 
     if index_dims:
         for name in index_dims:
@@ -330,9 +392,18 @@ def create_dims(ctx, dataframe, index_dims,
                 # in this case we use a scalar  (type-checked earlier)
                 dim_tile = tile
 
+            # get the FilterList, if any
+            if isinstance(filters, dict) and name in filters:
+                dim_filters = filters[name]
+            elif filters is not None:
+                dim_filters = filters
+            else:
+                dim_filters = None
+
             col = dataframe[name]
             dims.append(
-                dim_for_column(ctx, dataframe, col.values, name, tile=dim_tile)
+                dim_for_column(ctx, dataframe, col.values, name,
+                               tile=dim_tile, dim_filters=dim_filters)
             )
 
     return dims, sparse
@@ -388,6 +459,7 @@ def from_pandas(uri, dataframe, **kwargs):
     :return: None
 
     """
+    check_dataframe_deps()
     import pandas as pd
 
     if 'tiledb_args' in kwargs:
@@ -402,7 +474,8 @@ def from_pandas(uri, dataframe, **kwargs):
     sparse = tiledb_args['sparse']
     index_dims = tiledb_args.get('index_dims', None)
     mode = tiledb_args.get('mode', 'ingest')
-    attrs_filters = tiledb_args.get('attrs_filters', None)
+    attr_filters = tiledb_args.get('attr_filters', None)
+    dim_filters = tiledb_args.get('dim_filters', None)
     coords_filters = tiledb_args.get('coords_filters', None)
     full_domain = tiledb_args.get('full_domain', False)
     capacity = tiledb_args.get('capacity', False)
@@ -413,6 +486,9 @@ def from_pandas(uri, dataframe, **kwargs):
     date_spec = tiledb_args.get('date_spec', None)
     column_types = tiledb_args.get('column_types', None)
 
+    if mode != 'append' and tiledb.array_exists(uri):
+        raise TileDBError("Array URI '{}' already exists!".format(uri))
+
     write = True
     create_array = True
     if mode is not None:
@@ -420,9 +496,14 @@ def from_pandas(uri, dataframe, **kwargs):
             write = False
         elif mode == 'append':
             create_array = False
+            if not sparse and row_start_idx is None:
+                raise TileDBError("Cannot append to dense array without 'row_start_idx'")
         elif mode != 'ingest':
             raise TileDBError("Invalid mode specified ('{}')".format(mode))
 
+    if sparse == False and (index_dims is None or 'index_col' not in kwargs):
+        full_domain = True
+
     if capacity is None:
         capacity = 0 # this will use the libtiledb internal default
 
@@ -430,14 +511,6 @@ def from_pandas(uri, dataframe, **kwargs):
         ctx = tiledb.default_ctx()
 
     if create_array:
-        if attrs_filters is None:
-           attrs_filters = tiledb.FilterList(
-                [tiledb.ZstdFilter(1, ctx=ctx)])
-
-        if coords_filters is None:
-            coords_filters = tiledb.FilterList(
-                [tiledb.ZstdFilter(1, ctx=ctx)])
-
         if nrows:
             if full_domain is None:
                 full_domain = False
@@ -445,7 +518,7 @@ def from_pandas(uri, dataframe, **kwargs):
         # create the domain and attributes
         # if sparse==None then this function may return a default based on types
         dims, sparse = create_dims(ctx, dataframe, index_dims, sparse=sparse,
-                           tile=tile, full_domain=full_domain)
+                           tile=tile, full_domain=full_domain, filters=dim_filters)
 
         domain = tiledb.Domain(
            *dims,
@@ -454,9 +527,12 @@ def from_pandas(uri, dataframe, **kwargs):
 
         attrs, attr_metadata = attrs_from_df(dataframe,
                                              index_dims=index_dims,
-                                             filters=attrs_filters,
+                                             filters=attr_filters,
                                              column_types=column_types)
 
+        # don't set allows_duplicates=True for dense
+        allows_duplicates = allows_duplicates and sparse
+
         # now create the ArraySchema
         schema = tiledb.ArraySchema(
             domain=domain,
@@ -504,7 +580,16 @@ def from_pandas(uri, dataframe, **kwargs):
             if A.schema.sparse:
                 coords = []
                 for k in range(A.schema.ndim):
-                    coords.append(dataframe.index.get_level_values(k))
+                    dim_name = A.schema.domain.dim(k).name
+                    if not create_array and \
+                        dim_name not in dataframe.index.names and \
+                        dim_name != "__tiledb_rows":
+                        # this branch handles the situation where a user did not specify
+                        # index_col and is using mode='append'. We would like to try writing
+                        # with the columns corresponding to existing dimension name.
+                        coords.append(write_dict.pop(dim_name))
+                    else:
+                        coords.append(dataframe.index.get_level_values(k))
 
                 # TODO ensure correct col/dim ordering
                 A[tuple(coords)] = write_dict
@@ -573,6 +658,8 @@ def open_dataframe(uri, ctx=None):
     >>> tiledb.object_type("iris.tldb")
     'array'
     """
+    check_dataframe_deps()
+
     if ctx is None:
         ctx = tiledb.default_ctx()
     # TODO support `distributed=True` option?
@@ -640,7 +727,10 @@ def from_csv(uri, csv_file, **kwargs):
             * ``mode``: (default ``ingest``), Ingestion mode: ``ingest``, ``schema_only``,
               ``append``
             * ``full_domain``: Dimensions should be created with full range of the dtype
-            * ``attrs_filters``: FilterList to apply to all Attributes
+            * ``attr_filters``: FilterList to apply to Attributes: FilterList or Dict[str -> FilterList]
+                for any attribute(s). Unspecified attributes will use default.
+            * ``dim_filters``: FilterList to apply to Dimensions: FilterList or Dict[str -> FilterList]
+                for any dimensions(s). Unspecified dimensions will use default.
             * ``coords_filters``: FilterList to apply to all coordinates (Dimensions)
             * ``sparse``: (default True) Create sparse schema
             * ``tile``: Dimension tiling: accepts either Int or a list of Tuple[Int] with per-dimension
@@ -659,11 +749,8 @@ def from_csv(uri, csv_file, **kwargs):
     >>> tiledb.object_type("iris.tldb")
     'array'
     """
-    try:
-        import pandas
-    except ImportError as exc:
-        print("tiledb.from_csv requires pandas")
-        raise
+    check_dataframe_deps()
+    import pandas
 
     if 'tiledb_args' in kwargs:
         tiledb_args = kwargs.get('tiledb_args')
@@ -697,6 +784,9 @@ def from_csv(uri, csv_file, **kwargs):
         elif mode not in ['ingest', 'append']:
             raise TileDBError("Invalid mode specified ('{}')".format(mode))
 
+    if mode != 'append' and tiledb.array_exists(uri):
+        raise TileDBError("Array URI '{}' already exists!".format(uri))
+
     # this is a pandas pass-through argument, do not pop!
     chunksize = kwargs.get('chunksize', None)
 
