@@ -155,9 +155,10 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             A[s_ichars, times, cccc] = df.to_dict(orient='series')
 
         with tiledb.SparseArray(uri) as A:
-            df1 = pd.DataFrame.from_dict(A[:])
+            df_res = pd.DataFrame.from_dict(A[:])
             for col in df.columns:
-                assert_array_equal(df[col], df1[col])
+                # TileDB default return is unordered, so must sort to compare
+                assert_array_equal(df[col].sort_values(), df_res[col].sort_values())
 
     def test_dataframe_basic1(self):
         uri = self.path("dataframe_basic_rt1")
@@ -276,7 +277,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df_dict = df.to_dict(orient='series')
         df.set_index(['time', 'double_range'], inplace=True)
 
-        tiledb.from_dataframe(uri, df)
+        tiledb.from_dataframe(uri, df, sparse=True)
 
         with tiledb.open(uri) as A:
             ned_time = A.nonempty_domain()[0]
@@ -300,6 +301,11 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             # test .df[] indexing with query
             df_idx_res = A.query(attrs=['int_vals']).df[slice(*ned_time), :]
             tm.assert_frame_equal(df_idx_res, df)
+            # test .df[] with Arrow
+            df_idx_res = A.query(use_arrow=True).df[slice(*ned_time), :]
+            tm.assert_frame_equal(df_idx_res, df)
+            df_idx_res = A.query(use_arrow=False).df[slice(*ned_time), :]
+            tm.assert_frame_equal(df_idx_res, df)
 
     def test_csv_dense(self):
         col_size = 10
@@ -341,7 +347,10 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df.set_index(['time', 'double_range'], inplace=True)
 
         tmp_array = os.path.join(tmp_dir, "array")
-        tiledb.from_csv(tmp_array, tmp_csv, index_col=['time', 'double_range'], parse_dates=['time'])
+        tiledb.from_csv(tmp_array, tmp_csv,
+                        sparse=True,
+                        index_col=['time', 'double_range'],
+                        parse_dates=['time'])
 
         df_bk = tiledb.open_dataframe(tmp_array)
 
@@ -364,14 +373,20 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             tiledb.from_csv(tmp_array2a, tmp_csv2, index_col=['int_vals'], sparse=True, allows_duplicates=False)
 
         # try again, check from_csv(allows_duplicates=True, sparse=True)
-        tiledb.from_csv(tmp_array2b, tmp_csv2, index_col=['int_vals'],
-                        sparse=True, allows_duplicates=True, float_precision='round-trip')
+        tiledb.from_csv(tmp_array2b, tmp_csv2,
+                        index_col=['int_vals'],
+                        parse_dates=['time'],
+                        sparse=True, allows_duplicates=True,
+                        float_precision='round-trip')
 
         with tiledb.open(tmp_array2b) as A:
-            #self.assertTrue(A.schema.sparse)
-            res = A[:]
-            assert_array_equal(res['int_vals'], df.int_vals.values)
-            assert_array_almost_equal(res['double_range'], df.double_range.values)
+            self.assertTrue(A.schema.sparse)
+            res_df = A.df[:]
+            # the duplicate value is on the dimension and can be retrieved in arbitrary
+            # order. we need to re-sort in order to compare, to avoid spurious failures.
+            res_df.sort_values('time', inplace=True)
+            cmp_df = df.set_index('int_vals').sort_values(by='time')
+            tm.assert_frame_equal(res_df, cmp_df)
 
     def test_dataframe_csv_schema_only(self):
         col_size = 10
@@ -403,6 +418,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
                         parse_dates=['time'],
                         mode='schema_only',
                         capacity=1001,
+                        sparse=True,
                         tile={'time': 5},
                         coords_filters=coords_filters)
 
@@ -440,22 +456,22 @@ class PandasDataFrameRoundtrip(DiskTestCase):
                 A.schema.attr(0).filters[0].level, attrs_filters[0].level
             )
 
-        # Test mode='append'
-        tiledb.from_csv(tmp_array, tmp_csv,
-                        index_col=['time', 'double_range'], mode='append')
+        # Test mode='append' for from_csv
+        tiledb.from_csv(tmp_array, tmp_csv, mode='append', row_start_idx=0)
         df2 = make_dataframe_basic3(10, time_range=(t0, t1))
         df2.sort_values('time', inplace=True)
         df2.set_index(['time', 'double_range'], inplace=True)
-        tiledb.from_dataframe(tmp_array, df2, mode='append')
+
+        # Test mode='append' for from_pandas
+        tiledb.from_pandas(tmp_array, df2, row_start_idx=len(df2), mode='append')
 
         with tiledb.open(tmp_array) as A:
-            res = A[:]
-            df_bk = pd.DataFrame(res)
-            df_bk.set_index(['time','double_range'], inplace=True)
+            df_bk = A.df[:]
 
             df.set_index(['time','double_range'], inplace=True)
             df_combined = pd.concat([df, df2])
             df_combined.sort_index(level='time', inplace=True)
+            df_bk.sort_index(level='time', inplace=True)
             tm.assert_frame_equal(df_bk, df_combined)
 
     def test_dataframe_csv_chunked(self):
@@ -475,7 +491,8 @@ class PandasDataFrameRoundtrip(DiskTestCase):
                         index_col=['double_range'],
                         parse_dates=['time'],
                         date_spec={'time': "%Y-%m-%dT%H:%M:%S.%f"},
-                        chunksize=10)
+                        chunksize=10,
+                        sparse=True)
 
         with tiledb.open(tmp_array) as A:
             res = A[:]
@@ -506,7 +523,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
 
             # test .df[] indexing
             df_idx_res = A.df[int(ned[0]):int(ned[1])]
-            tm.assert_frame_equal(df_idx_res, df)
+            tm.assert_frame_equal(df_idx_res, df, check_index_type=False)
 
             # test .df[] indexing with query
             df_idx_res = A.query(attrs=['time']).df[int(ned[0]):int(ned[1])]
@@ -515,6 +532,12 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             df_idx_res = A.query(attrs=['double_range']).df[int(ned[0]):int(ned[1])]
             tm.assert_frame_equal(df_idx_res, df[['double_range']])
 
+            # test .df[] indexing with arrow
+            df_idx_res = A.query(use_arrow=True, attrs=['time']).df[int(ned[0]):int(ned[1])]
+            tm.assert_frame_equal(df_idx_res, df[['time']])
+            df_idx_res = A.query(use_arrow=True, attrs=['double_range']).df[int(ned[0]):int(ned[1])]
+            tm.assert_frame_equal(df_idx_res, df[['double_range']])
+
             # disable coordinate dimension/index
             df_idx_res = A.query(coords=False).df[int(ned[0]):int(ned[1])]
             tm.assert_frame_equal(df_idx_res, df.reset_index(drop=True))
@@ -532,7 +555,8 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df.to_csv(tmp_csv, index=False, na_rep="NaN")
 
         tmp_array = os.path.join(tmp_dir, "array")
-        tiledb.from_csv(tmp_array, tmp_csv, fillna={'v': 0})
+        # TODO: test Dense too
+        tiledb.from_csv(tmp_array, tmp_csv, fillna={'v': 0}, sparse=True)
 
         def check_array(path, df):
             # update the value in the original dataframe to match what we expect on read-back
@@ -540,16 +564,18 @@ class PandasDataFrameRoundtrip(DiskTestCase):
 
             with tiledb.open(path) as A:
                 df_bk = A.df[:]
-                tm.assert_frame_equal(df_bk, df)
+                tm.assert_frame_equal(df_bk, df, check_index_type=False)
 
         check_array(tmp_array, copy.deepcopy(df))
 
+        # Test writing a StringDtype in newer pandas versions
         if hasattr(pd, 'StringDtype'):
             tmp_array2 = os.path.join(tmp_dir, "array2")
             tiledb.from_csv(tmp_array2,
                             tmp_csv,
                             fillna={'v': 0},
-                            column_types={'v': pd.Int64Dtype})
+                            column_types={'v': pd.Int64Dtype},
+                            sparse=True)
             df_to_check = copy.deepcopy(df)
             df_to_check['v'][4] = 0
             df_to_check = df_to_check.astype({'v': np.int64})
@@ -592,8 +618,83 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df_orig = pd.concat(input_dfs, axis=0).set_index(["time"]).sort_values('time')
 
         with tiledb.open(tmp_array) as A:
-            res = A[:]
-            df_bk = pd.DataFrame(res)
-            df_bk = df_bk.set_index(['time'])
+            df_bk = A.df[:]
+            # TileDB default return is unordered, so sort to compare
+            df_bk = df_bk.sort_index()
 
             tm.assert_frame_equal(df_bk, df_orig)
+
+    def test_dataframe_misc(self):
+        uri = self.path("test_small_domain_range")
+        df = pd.DataFrame({'data': [2]}, index=[0])
+        tiledb.from_pandas(uri, df)
+
+        data = {'data': np.array([1,2,3]),
+                'raw': np.array([4,5,6]),
+                'index': np.array(['a', 'b', 'c'], dtype=np.dtype('|S')),
+                'indey': np.array([0.0,0.5,0.9])}
+        df = pd.DataFrame.from_dict(data)
+        df = df.set_index(['index', 'indey'])
+        uri = self.path("test_string_index_infer")
+        tiledb.from_pandas(uri, df)
+        with tiledb.open(uri) as A:
+            self.assertTrue(A.schema.domain.dim(0).dtype == np.dtype('|S'))
+
+        # test setting Attr and Dim filter list by override
+        uri = self.path("test_df_attrs_filters1")
+        bz_filter = [tiledb.Bzip2Filter(4)]
+        def_filter = [tiledb.GzipFilter(-1)]
+        tiledb.from_pandas(uri, df, attr_filters=bz_filter, dim_filters=bz_filter)
+        with tiledb.open(uri) as A:
+            self.assertTrue(A.schema.attr('data').filters == bz_filter)
+            self.assertTrue(A.schema.attr('raw').filters == bz_filter)
+            self.assertTrue(A.schema.domain.dim('index').filters == bz_filter)
+            self.assertTrue(A.schema.domain.dim('indey').filters == bz_filter)
+
+        # test setting Attr and Dim filter list by dict
+        uri = self.path("test_df_attrs_filters2")
+        tiledb.from_pandas(uri, df,
+                           attr_filters={'data': bz_filter},
+                           dim_filters={'index': bz_filter})
+        with tiledb.open(uri) as A:
+            self.assertTrue(A.schema.attr('data').filters == bz_filter)
+            self.assertTrue(A.schema.attr('raw').filters == tiledb.FilterList())
+            self.assertTrue(A.schema.domain.dim('index').filters == bz_filter)
+            self.assertTrue(A.schema.domain.dim('indey').filters == tiledb.FilterList())
+
+    def test_dataframe_query(self):
+        uri = self.path("df_query")
+
+        col_size = 10
+        df = make_dataframe_basic3(col_size)
+        df.set_index(['time'], inplace=True)
+
+        tiledb.from_dataframe(uri, df, sparse=True)
+
+        with tiledb.open(uri) as A:
+            with self.assertRaises(tiledb.TileDBError):
+                A.query(dims=['nodimnodim'])
+            with self.assertRaises(tiledb.TileDBError):
+                A.query(attrs=['noattrnoattr'])
+
+            res_df = A.query(dims=['time'], attrs=['int_vals']).df[:]
+            self.assertTrue('time' == res_df.index.name)
+            self.assertTrue('int_vals' in res_df)
+            self.assertTrue('double_range' not in res_df)
+
+            # try index_col alone: should have *only* the default RangeIndex column
+            res_df2 = A.query(index_col=None).df[:]
+            self.assertTrue(isinstance(res_df2.index, pd.RangeIndex))
+
+            # try no dims, index_col None: should only value cols and default index
+            res_df3 = A.query(dims=False, index_col=None).df[:]
+            self.assertTrue('time' not in res_df3)
+            self.assertTrue('int_vals' in res_df3)
+            self.assertTrue('double_range' in res_df3)
+            self.assertTrue(isinstance(res_df3.index, pd.RangeIndex))
+
+            # try attr as index_col:
+            res_df4 = A.query(dims=False, index_col=['int_vals']).df[:]
+            self.assertTrue('time' not in res_df4)
+            self.assertTrue('double_range' in res_df4)
+            self.assertTrue('int_vals' == res_df4.index.name)
\ No newline at end of file
